#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include "constants.h"
#include "types.h"
#include "ibdb.h"
#include "mem_manage.h"


void
discard_sequence()
{
}

/***************************************************************/
/*
 * extended link control frames -- like PLOGI/FLOGI and LOGO
 *
 * Input Parameters: pointers to the IBDB, and RCTL
 * Output Parameters: none
 */
/***************************************************************/
void
extended_link( IBDB *Ptr_IBDB, unsigned int R_CTL )
{
  ESCB *New_ESCB;
  unsigned int Parm;
  unsigned int S_ID;

  Parm = Ptr_IBDB->DA[0];

  if ( Parm == 0x03000000 )
    {
      /* PLOGI login frame */
      /* check if alredy logged-in */
      S_ID = Ptr_IBDB->FH.S_ID;

      /* if not alredy logged-in, open a new session */
      
      /* open a new ESCB and SSCB for the login frame */
      
      /* generate ACK */

      /* open a new sequence and send an ACC frame if successful
         login, LS_RJT if not */
      

      /* make new ESCB */
      New_ESCB = get_ESCB();

      New_ESCB->OX_ID = Ptr_IBDB->FH.OX_ID;
      New_ESCB->RX_ID = get_ESCB() - ESCB_memory;
      New_ESCB->D_ID = Ptr_IBDB->FH.D_ID;

      /*      wordptr = (unsigned int *) & Ptr_IBDB->DA; */
      New_ESCB->EE_CREDIT = (Ptr_IBDB->DA[11] && 0x0000FFFF) ;

      
    }

  return;
}

/***************************************************************/
/*
 * generate acknowledgement for the received frame for class 1 and
 * class 2 services
 *
 * Input Parameters: pointers to the IBDB, ESCB and SSCB, indication
 *                   for the for the first frame
 * Output Parameters: none
 *  
 */
/***************************************************************/
void
generate_ack( IBDB *Ptr_IBDB, ESCB *Current_ESCB, SSCB *Current_SSCB, 
	      unsigned int F_CTL, unsigned int Source_ID, unsigned int OX_ID, 
	      unsigned int SOFr, IBDB **MinOFC, IBDB **MaxOFC, 
	      IBDB **MinOIB, IBDB **MaxOIB)
{
  OBDB *Pointer_OBDB;
  IBDB **OFC_head;

  IBDB *First_IBDB;
  IBDB **OIB_head;

  unsigned int SEQ_CNT, Highest_SEQ_CNT, Frame_CNT, Lowest_SEQ_CNT;
  unsigned int test2, test1;
  unsigned int EOFindicate;

  EOFindicate = 0;

  SEQ_CNT = Ptr_IBDB->FH.SEQ_CNT;
  Highest_SEQ_CNT = Current_SSCB->Highest_SEQ_CNT;
  Lowest_SEQ_CNT = Current_SSCB->Lowest_SEQ_CNT;
  Frame_CNT = Current_SSCB->Frame_CNT;

  /* Not all frames received */
  test1 = (Frame_CNT != (SEQ_CNT - Lowest_SEQ_CNT + 1));
  test2 = (Frame_CNT != (Highest_SEQ_CNT - Lowest_SEQ_CNT + 1));

  if ((test1 && Highest_SEQ_CNT == 0) || (test2 && Highest_SEQ_CNT != 0))
    {
      /* if not all frames received, do not generate ACK */
      return;
    }

  /* end of sequence detected and all frames received. */
  /* generate ACK with the highest SEQ_CNT */
  test2 = (Highest_SEQ_CNT != 0) && 
    (Frame_CNT == (Highest_SEQ_CNT - Lowest_SEQ_CNT + 1));
  if (test2)
    {
      SEQ_CNT = Highest_SEQ_CNT;
      EOFindicate = 1;
    }

  /* allocate an OBDBs block for the acknowledgement frame */
  Pointer_OBDB = get_OBDB();

  /* assemble the acknowledgement frame */
  /* interchange D_ID and S_ID of the received frame */
  Pointer_OBDB->FH.D_ID = Source_ID;
  Pointer_OBDB->FH.S_ID = Ptr_IBDB->FH.D_ID;

  /* several fields stay the same */
  Pointer_OBDB->FH.OX_ID = OX_ID;
  Pointer_OBDB->FH.SEQ_ID = Ptr_IBDB->FH.SEQ_ID;
  Pointer_OBDB->FH.TYPE = Ptr_IBDB->FH.TYPE;
  Pointer_OBDB->FH.DF_CTL = 0;

  /* write SEQ_CNT and RX_ID  */
  Pointer_OBDB->FH.SEQ_CNT = SEQ_CNT;
  Pointer_OBDB->FH.RX_ID = Current_ESCB->RX_ID;

  /* invert the sequence and exchange context bits in the F_CTL field */
  Pointer_OBDB->FH.F_CTL = F_CTL ^ (3<<22);

  /* put ACK code in the R_CTL field. 0xC0 is for ACK_1, 0xC1 is for
     ACK_0 and ACK_N.  Which ACK form to use depends on the contents
     of the F_CTL field. */
  switch (F_CTL & (3<<12))
    {
    case 0:                 /* use ACK_1 */
    case 1:
      Pointer_OBDB->FH.R_CTL = 0xC0;
      Pointer_OBDB->FH.FRAME_HEADER_PARAMETER = 1;
      break;
      
    case 2:                 /* use ACK_N */
      Pointer_OBDB->FH.R_CTL = 0xC1;
      Pointer_OBDB->FH.FRAME_HEADER_PARAMETER = 1;
      break;

    case 3:                 /* ACK_0 required */
      Pointer_OBDB->FH.R_CTL = 0xC1;
      Pointer_OBDB->FH.FRAME_HEADER_PARAMETER = 0;
      break;
    }

  /* provide SOF and EOF delimiters for the ACK frame */  
  switch ( SOFr )
    {
    case SOFc1:                    /* class 1 frames */
    case SOFi1:
    case SOFn1:           
      Pointer_OBDB->SOFr = SOFn1;
      if ( EOFindicate )      
	Pointer_OBDB->EOFr = EOFdt;
      else
	Pointer_OBDB->EOFr = EOFn; 
      break;
    case SOFi2:                    /* class 2 frames */
    case SOFn2:
      Pointer_OBDB->SOFr = SOFn2;
      if ( EOFindicate )      
	Pointer_OBDB->EOFr = EOFt;
      else
	Pointer_OBDB->EOFr = EOFn;
      break;
    }
  
  /* write the length of the data for this frame (it is 0) */
  Pointer_OBDB->Frame_Length = 0;

  /* after all data are prepared, put the OBDB data block into
     outbound FC queue */

  /* write OBDB address into ring 5 */
  OFC_head = Ring_buffer_pointer_table[10];

  put_ring((IBDB *)Pointer_OBDB, OFC_head, MinOFC, MaxOFC, 5);

  if (test2)
    {
      /* last ACK, put the sequence in the IB outgoing ring buffer 6 */
      /* find the first IBDB in the first chain of the sequence */
      First_IBDB = Current_SSCB->First_IBDB_Pointer;
      OIB_head = Ring_buffer_pointer_table[12];
      
      put_ring(First_IBDB, OIB_head, MinOIB, MaxOIB, 6);
    }

}

/***************************************************************/
/*
 * chain the received frame in a frame linked list
 * if the frame is out of order, reorder frames
 *
 * Input Parameters: pointers to the IBDB and SSCB
 * Output Parameters: none
 *
 */
/***************************************************************/
unsigned int
frame_ordering ( IBDB *Ptr_IBDB, SSCB *Current_SSCB, unsigned int F_CTL)
{
  IBDB *Previous_IBDB;
  unsigned int SOFr;
  unsigned int Highest_SEQ_CNT, SEQ_CNT, Expected_SEQ_CNT, Frame_CNT;

  SOFr = Ptr_IBDB->BD.Frame_Summary.SOFr;
  SEQ_CNT = Ptr_IBDB->FH.SEQ_CNT;

  Frame_CNT = Current_SSCB->Frame_CNT;
  Highest_SEQ_CNT = Current_SSCB->Highest_SEQ_CNT;
  Expected_SEQ_CNT = Current_SSCB->Expected_SEQ_CNT;


  /* Initialize for the first frame of the sequence? */
  if ( (SOFr==SOFi1) || (SOFr==SOFi2) || (SOFr == SOFi3) )
    {
      /* data needed for frame chaining */
      Current_SSCB->Previous_IBDB_Pointer = Ptr_IBDB;
      Current_SSCB->Lowest_SEQ_CNT = SEQ_CNT;
      Highest_SEQ_CNT = 0;
      Current_SSCB->First_IBDB_Pointer = Ptr_IBDB;
      
      /* data needed for frame reordering */
      Frame_CNT = 0;
      Expected_SEQ_CNT = SEQ_CNT;
      Current_SSCB->Previous_Block_IBDB_Pointer = Ptr_IBDB; 
      Current_SSCB->Previous_SEQ_CNT = SEQ_CNT;
      
      /* data needed for end to end credit */
      Current_SSCB->Lower_SEQ_CNT = SEQ_CNT;
    }

  else if ( F_CTL & (1<<19) )
    {
      /* Last frame in the sequence */
      Highest_SEQ_CNT = SEQ_CNT;
      /* the next frame pointer points to NULL */
      Ptr_IBDB->BD.PTR_NEXT_DB = NULL;
    }

  ++Frame_CNT;
  

  if ( (SEQ_CNT == Expected_SEQ_CNT) || (SOFr == SOFn3) )
    {
      /* The frame is received in order, or is in class 3
         service. Chain data frames. */
      if ( Frame_CNT != 1)
	{  
	  /* 1st frame is not chained */
	  Previous_IBDB = Current_SSCB->Previous_IBDB_Pointer;
	  Previous_IBDB->BD.PTR_NEXT_DB = Ptr_IBDB;      
	  Current_SSCB->Previous_IBDB_Pointer = Ptr_IBDB;
	}
      ++Expected_SEQ_CNT;
    }
  else
  {
    /* Received frame is out-of-order */
    /* For class 1 service, the sequence is discarded, exept for the
       first frame in the sequence. */
    if ((SOFr == SOFn1) && (Frame_CNT != 1))
      {
	discard_sequence();
      }
    else
      {
	/* for class 2 service, record the frame chain received in
	   order in the reassembly table, and update the fields in the
	   SSCB. */
	if ( Frame_CNT != 1 )
	  {
	    hashinsert_RA( Current_SSCB->Previous_SEQ_CNT,
			   Current_SSCB->Previous_Block_IBDB_Pointer );
	  }
	Current_SSCB->Previous_IBDB_Pointer = Ptr_IBDB;
	Current_SSCB->Previous_Block_IBDB_Pointer = Ptr_IBDB; 
	Current_SSCB->Previous_SEQ_CNT = SEQ_CNT;

	Expected_SEQ_CNT = ++SEQ_CNT;

	/* as this is the last frame in the chain, point it to null */
	Ptr_IBDB->BD.PTR_NEXT_DB = NULL;
      }
  }

  /*  Record last frame received */
  if ( (Highest_SEQ_CNT != 0) &&
       (Frame_CNT == (Highest_SEQ_CNT - Current_SSCB->Lowest_SEQ_CNT + 1)) )
    {
      hashinsert_RA( Current_SSCB->Previous_SEQ_CNT,
		     Current_SSCB->Previous_Block_IBDB_Pointer );
    }

  Current_SSCB->Expected_SEQ_CNT = Expected_SEQ_CNT;
  Current_SSCB->Frame_CNT =Frame_CNT;
  Current_SSCB->Highest_SEQ_CNT = Highest_SEQ_CNT;

  return SOFr;
}

/***************************************************************/
/*
 * Inmpements end-to-end flow control protocol for class 1 and class 2
 * services.  The credit is incremented for each sent frame, and
 * decreased with each received ACK.
 */
/***************************************************************/
/*
 * Increase EE_credit.  Called by sequence initiator each time a frame
 * has been sent.  
 * 
 * Input Parameters: none
 * Output Parameters: none 
 */
/***************************************************************/
void
EE_increase( IBDB *Ptr_IBDB, SSCB *Current_SSCB )
{
  int SOFr;

  SOFr = Ptr_IBDB->BD.Frame_Summary.SOFr;

  /* initialize EE credit for establishing connection in class 1 SOFc1 */
  if ( SOFr == SOFc1 )
    Current_SSCB->EE_CNT = 0;   
  /* another notation Current_ESCB->EE_CREDIT; */


  /* EE credit is incremented each time class 2 or class 1 frame is
     sent. */
  if (  (SOFr == SOFc1) || (SOFr == SOFi1) || (SOFr == SOFn1) || 
	(SOFr == SOFi2) || (SOFr == SOFn2) )
    ++Current_SSCB->EE_CNT;

  return;
}

/***************************************************************/
/*
 * Decrease EE_credit each time ACK, BSY or RJT frames are received
 * 
 * Input Parameters: none
 * Output Parameters: none 
 */
/***************************************************************/
void
EE_credit( IBDB *Ptr_IBDB, ESCB *Current_ESCB, SSCB *Current_SSCB,
	   unsigned int R_CTL)
{
  int ACK_CNT, EE_CNT, SEQ_CNT;
  unsigned int Parm;

  EE_CNT = Current_SSCB->EE_CNT;
  Parm = Ptr_IBDB->FH.FRAME_HEADER_PARAMETER;
  SEQ_CNT = Ptr_IBDB->FH.SEQ_CNT;

  /* decrement credit when P_RJT, P_BSY, F_RJT or F_BSY to data frame
     received */
  if ( (R_CTL == 0xC2) || (R_CTL == 0xC3) || (R_CTL == 0xC4) || 
       (R_CTL == 0xC5) )
    --EE_CNT;
  
  /* decrease credit when ACK_1 or ACK_N with history bit 1 received */
  if ( ((R_CTL == 0xC0) || (R_CTL == 0xC1)) && (Parm & (1<<16)) )
    {
      ACK_CNT = Parm & 0x0000ffff;
      EE_CNT -= ACK_CNT;
      Current_SSCB->Missing_ACK += ACK_CNT; 
    }

  /* decrease credit if ACK_1 or ACK_N with history bit 0 received for
     the number of missing ACKs */
  if ( ((R_CTL == 0xC0) || (R_CTL == 0xC1)) && ((Parm & (1<<16)) == 0) )
    {
      /* do not adjust credit if lower SEQ_CNT arrives later */
      if ( SEQ_CNT < Current_SSCB->Lower_SEQ_CNT )
	return;
    }
  else
    {
      ACK_CNT = Parm & 0x0000ffff;
      EE_CNT -= 
	(SEQ_CNT - Current_SSCB->Lower_SEQ_CNT 
	 + Current_SSCB->Missing_ACK);
      Current_SSCB->Lower_SEQ_CNT = SEQ_CNT;
      Current_SSCB->Missing_ACK = 0; 
    }

  /* credit can not be negative */
  if ( EE_CNT <= 0 )
    EE_CNT = 0;

  Current_SSCB->EE_CNT = EE_CNT;
}

/***************************************************************/
/*
 * retrieve the exchange for the received frame 
 * 
 * find matching ESCB (exchange status and control block)
 * and make it current exchange
 *
 * Input Parameters: pointer to the IBDB, references to ESCB and SSCB
 * Output Parameters: pointers to the ESCB and the SSCB in supplied
 * addresses, indicaton for X_ID interlock, indicator FirstFrame of an
 * exchange, a new RX_ID has been assigned. 
 */
/***************************************************************/
void
context_switch(IBDB *Ptr_IBDB, unsigned int Source_ID, unsigned int OX_ID, 
	       ESCB **found_ESCB, SSCB **found_SSCB, unsigned int F_CTL)
{

  ESCB *Current_ESCB;  
  SSCB *Current_SSCB;  

  unsigned int ExchangeIndex;
  unsigned int SequenceIndex;
  unsigned int FirstFrame;

  unsigned int KEY;
  unsigned int RX_ID, SEQ_ID;

  int i;
 
  ExchangeIndex = 0xFFFF;
  FirstFrame = 0;

  if ( (F_CTL & (1<<23)) == 0 )
    /* Exchange originator, find ESCB using OX_ID! */
    ExchangeIndex = OX_ID;
  
  else 
    {
      /* Exchange responder. Is RX_ID already assigned? */
      RX_ID = Ptr_IBDB->FH.RX_ID; 

      if (RX_ID != 0xFFFF)		
	ExchangeIndex = RX_ID;

      /* find ESCB block using RX_ID In the future, we may assemble an
	 address with some shifts etc., but this requires to
	 preallocate memory pages at 2^23 byte boundary and is quite
	 unportable... */

      else
	{
	  /* RX_ID and ESCB already assigned? Use S_ID and OX_ID for
	     identification. */
	  
	  KEY = ((((Source_ID >> 4) ^ OX_ID) >> 6) & 0x0000000F);

	  ExchangeIndex = hashsearch_RX(OX_ID, Source_ID, KEY);
	  
	  /* if RX_ID not in the table */
	  if ( ExchangeIndex == 0xFFFF )
	    {
	      /* New exchange, assign new RX_ID, new ESCB and make a
		 new entry in the RXmaptable!  */

	      ExchangeIndex = get_ESCB() - ESCB_memory;
	      
    	      hashinsert_RX(OX_ID, Source_ID, ExchangeIndex, KEY);

	      FirstFrame = 1;
	    }
	}
    }

  assert (ExchangeIndex < MAX_ESCB);
  Current_ESCB = & ESCB_memory[ExchangeIndex];

  /* write the OX_ID and RX_ID fileds in the newly opened ESCB */
  if ( FirstFrame )
    {
      Current_ESCB->OX_ID = OX_ID; 
      Current_ESCB->RX_ID = ExchangeIndex;
      Current_ESCB->D_ID = Source_ID;
    }      
  

  *found_ESCB = Current_ESCB;

  /*************************************************************/
  /* check does the received frame belong to one of 4 recepient
     sequences or 2 initiated sequences.  From the active ESCB block
     use sequence indexes SEQ_IND[0...3], if sequence recipient, or
     SEQ_IND[45] if sequence originator. */

  SequenceIndex = 0;
  SEQ_ID = Ptr_IBDB->FH.SEQ_ID;

  if (F_CTL & (1<<22))
    {
      /* sequence recepient, we have probably received a data frame */

      if ( FirstFrame )
	{
	  /* Seq.ID is not in the table */
	  SequenceIndex = 0; 
	}
      else
	for (i=0; i<=3; i++)
	  {
	    if ( SEQ_ID == Current_ESCB->SEQ_IND[i].SEQ_ID ) 
	      SequenceIndex = Current_ESCB->SEQ_IND[i].SSCB_Index;  
	    break;
	  }
      
      if ( SequenceIndex == 0)
	{
	  /* the sequence number is not in the ESCB. Can the oldest
	     recipient SEQ_ID (in SEQ_IND[3]) be retired?  (if
	     sequence closed and inactive?) */
	  if ( Current_ESCB->SEQ_IND[3].S_PARM  & (1<<31) )
	    {
	      /* sequence open, generate error - too many open sequences */
	    }
	  else
	    /* the oldest sequence is closed, shift it out.  Make a
	       new entry for the sequence in SEQ_IND[0]. */
	    {
	      for (i=2; i>=0; i--)
		Current_ESCB->SEQ_IND[i+1] = Current_ESCB->SEQ_IND[i]; 

	      /* get an address of the next free SSCB using get_SSCB.
		 Transfer the address into the array index for SSCB
		 memory*/
	      SequenceIndex = get_SSCB() - SSCB_memory;

	      /* Make a new entry for the sequence in SEQ_IND[0]. */	      
	      Current_ESCB->SEQ_IND[0].SEQ_ID = SEQ_ID; 
	      Current_ESCB->SEQ_IND[0].SSCB_Index = SequenceIndex;
	      Current_ESCB->SEQ_IND[0].S_PARM = 3 << 30; 
	    }
	}       
    } 
  else
    /* sequence initiator, we have probably received a control frame
       like ACK.  SequenceIndex is already assigned. */
    {
 
      /* sequence initiator.  SequenceIndex is already assigned. */
      if (SEQ_ID == Current_ESCB->SEQ_IND[4].SEQ_ID) 
	SequenceIndex = Current_ESCB->SEQ_IND[4].SSCB_Index;  
      else
	{
	  if (SEQ_ID == Current_ESCB->SEQ_IND[5].SEQ_ID) 
	    SequenceIndex = Current_ESCB->SEQ_IND[5].SSCB_Index;  
	}
    }

  assert (SequenceIndex < MAX_SSCB);
  Current_SSCB = & SSCB_memory[SequenceIndex];

  *found_SSCB = Current_SSCB;

}
  
/***************************************************************/
/*
 * Once the data have been moved into the memory,
 * FC protocol tasks on the inbound frames are implemented.
 */
/***************************************************************/
int
main()
{
  unsigned int SOFr = 0;

  unsigned int i, k;
  unsigned int test1, test2;
  unsigned int word1, word2, word3, word4;

  IBDB *Ptr_IBDB;
  ESCB *Current_ESCB;
  SSCB *Current_SSCB;
  IBDB **Local_tail, **Local_head, **LMin, **LMax;

  IBDB **R0_head, **Min0, **Max0, **R1_head, **Min1, **Max1;
  IBDB **R2_head, **Min2, **Max2, **R3_head, **Min3, **Max3; 
  IBDB **Master_tail, **Master_head, **Min, **Max;
  IBDB **MinOIB, **MaxOIB;
  IBDB **MinOFC, **MaxOFC;
  IBDB **MinIIB, **MaxIIB;
  
  unsigned int Source_ID, OX_ID, F_CTL, R_CTL;

  /* initialize master ring buffer pointers */

  Min = & Ring_buffer[0];
  Max = & Ring_buffer[MAX_RING_MASTER - 1];

  /* initialize local ring buffer pointers */

  Min0 = & Ring_buffer_L0[0];
  Max0 = & Ring_buffer_L0[MAX_RING_LOCAL - 1];
  Ring_buffer_pointer_table[2] = Min0;
  Ring_buffer_pointer_table[3] = Max0;

  Min1 = & Ring_buffer_L1[0];
  Max1 = & Ring_buffer_L1[MAX_RING_LOCAL - 1];
  Ring_buffer_pointer_table[4] = Min1;
  Ring_buffer_pointer_table[5] = Max1;

  Min2 = & Ring_buffer_L2[0];
  Max2 = & Ring_buffer_L2[MAX_RING_LOCAL - 1];
  Ring_buffer_pointer_table[6] = Min2;
  Ring_buffer_pointer_table[7] = Max2;

  Min3 = & Ring_buffer_L3[0];
  Max3 = & Ring_buffer_L3[MAX_RING_LOCAL - 1];
  Ring_buffer_pointer_table[8] = Min3;
  Ring_buffer_pointer_table[9] = Max3;

  /* initialize FC output ring buffer pointers */

  MinOFC = (IBDB **) & Ring_buffer_OFC[0];
  MaxOFC = (IBDB **) & Ring_buffer_OFC[MAX_RING_OFC - 1];
  Ring_buffer_pointer_table[10] = MinOFC;
  Ring_buffer_pointer_table[11] = MaxOFC;

  /* initialize IB output ring buffer pointers */

  MinOIB = & Ring_buffer_OIB[0];
  MaxOIB = & Ring_buffer_OIB[MAX_RING_OIB - 1];
  Ring_buffer_pointer_table[12] = MinOIB;
  Ring_buffer_pointer_table[13] = MaxOIB;

  /* initialize IB input ring buffer pointers */

  MinIIB = (IBDB **) & Ring_buffer_IIB[0];
  MaxIIB = (IBDB **) & Ring_buffer_IIB[MAX_RING_IIB - 1];
  Ring_buffer_pointer_table[14] = MinIIB;
  Ring_buffer_pointer_table[15] = MaxIIB;

  initialize_OBDB();
  initialize_ESCB();
  initialize_SSCB();

  /* initialize registers */
  Source_ID = 0;
  OX_ID = 0;

  printf("initialization done\n");

  for (i=0;i<=4;i++)
    {  
      /* read an entry from the master input queue */
      Master_tail = Ring_buffer_pointer_table[1];
 
      Ptr_IBDB = get_ring(Master_tail, Min, Max, 0);

      /* read data from frame header required for hash function */

      /* get S_ID */
      word1 = Ptr_IBDB->FH.S_ID;
      word2 = word1 & 0x00000FFF;
      word1 = (word1 >> 12) & 0x00000FFF;

      /* get OX_ID */
      word3 = Ptr_IBDB->FH.OX_ID;
      word4 = word3 & 0x000000FF; 
      word3 = (word3 >> 8) & 0x000000FF;

      /* calculate hash function */
      word1 += word3;
      word2 += word4;
      word1 ^= word2;

      word3 = word1 & 0x0000003F;
      word1 >>= 6;
      word1 += word3;

      word3 = word1 & 0x0000000F;
      word1 >>= 4;
      word1 ^= word3;

      word3 = word1 & 0x00000003;
      word1 >>= 2;
      word1 ^= word3;

      /* write the IBDB to one of the four local ring buffers */
      switch (word1)
	{
	case 0:                 /* thread 1 */
	  /* write IBDB address into ring L0 */
	  R0_head = Ring_buffer_pointer_table[2];

	  put_ring(Ptr_IBDB, R0_head, Min0, Max0, 1);

	  k = 1;
	  LMin = & Ring_buffer_L0[0];
	  LMax = & Ring_buffer_L0[MAX_RING_LOCAL - 1];
	  break; 

	case 1:                 /* thread 2 */
	  /* write IBDB address into ring L1 */
	  R1_head = Ring_buffer_pointer_table[4];

	  put_ring(Ptr_IBDB, R1_head, Min1, Max1, 2);

	  k = 2;
	  LMin = & Ring_buffer_L1[0];
	  LMax = & Ring_buffer_L1[MAX_RING_LOCAL - 1];
	  break; 
	
	case 2:                /* thread 3 */
	  /* write IBDB address into ring L2 */
	  R2_head = Ring_buffer_pointer_table[6];

	  put_ring(Ptr_IBDB, R2_head, Min2, Max2, 3);

	  k = 3;
	  LMin = & Ring_buffer_L2[0];
	  LMax = & Ring_buffer_L2[MAX_RING_LOCAL - 1];
	  break; 
	
	case 3:                 /* thread 4 */
	  /* write IBDB address into ring L3 */
	  R3_head = Ring_buffer_pointer_table[8];

	  put_ring(Ptr_IBDB, R3_head, Min3, Max3, 4);

	  k = 4;
	  LMin = & Ring_buffer_L3[0];
	  LMax = & Ring_buffer_L3[MAX_RING_LOCAL - 1];      
	  break; 
	}     
    }      
  
  for (i = 0; i<=4; i++)
    { 
      /* read an entry from the local input queue */
      Local_tail = Ring_buffer_pointer_table[2*k + 1];
      Local_head = Ring_buffer_pointer_table[2*k];
      
      Ptr_IBDB = get_ring(Local_tail, Min, Max, k);


      /****************************************************************/  
      /* Is the received frame a data or a link control frame? */

      R_CTL = Ptr_IBDB->FH.R_CTL;
      
      if ( R_CTL == 0x22 )
	/* This is an extended link control frame, like login */
	{
	  extended_link( Ptr_IBDB, R_CTL );
	}
      else
	{
	  /********************************************************/
	  /* Context switch: Find the exchange status and control
	     block and the sequence status and control block of the
	     received frame. */

	  /* test if current exchange is same as last exchange and
	     bail out early */
	  test1 = Ptr_IBDB->FH.S_ID;
	  test2 = Ptr_IBDB->FH.OX_ID;
	      
	  F_CTL = Ptr_IBDB->FH.F_CTL;
	  
	  if ( (Source_ID != test1) || (OX_ID != test2) )
	    { 
	      /* current exchange is different from the last exchange */ 
	      Source_ID = test1;
	      OX_ID = test2;
	      
	      context_switch(Ptr_IBDB, Source_ID, OX_ID, &Current_ESCB, 
			     &Current_SSCB, F_CTL);
	    }

	  if ( (R_CTL & 0xF0) == 0xC0 )
	    /* This is a link control frame, update EE_credit */
	    {
	      EE_credit( Ptr_IBDB, Current_ESCB, Current_SSCB, R_CTL );
	    }
	  else
	    {
	      /* This is data frame */
	      /* chain the received frame in a linked frame list */

	      SOFr = frame_ordering(Ptr_IBDB, Current_SSCB, F_CTL); 
	      
	      /* Does the frame require acknowledgement to be
		 generated?  Generate an acknowledgement for the class
		 1 and 2 services. */
	      
	      if ((SOFr==SOFi1) || (SOFr==SOFn1) || (SOFr==SOFi2) || 
		  (SOFr==SOFn2))
		generate_ack( Ptr_IBDB, Current_ESCB, Current_SSCB, 
			      F_CTL, Source_ID, OX_ID, SOFr,
			      MinOFC, MaxOFC, MinOIB, MaxOIB);
	    }
	}
    }
  return 0;
}




